/*######     Copyright (c) 1997-2015 Ufasoft  http://ufasoft.com  mailto:support@ufasoft.com,  Sergey Pavlov  mailto:dev@ufasoft.com #########################################################################################################
#                                                                                                                                                                                                                                            #
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation;  either version 3, or (at your option) any later version.          #
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.   #
# You should have received a copy of the GNU General Public License along with this program; If not, see <http://www.gnu.org/licenses/>                                                                                                      #
############################################################################################################################################################################################################################################*/

#include <el/ext.h>

#include "prime.h"

#include EXT_HEADER_DYNAMIC_BITSET

namespace Ext { namespace Num {

class PrimeNumbers : public vector<uint32_t> {
	typedef dynamic_bitset<> base;
public:
	PrimeNumbers() {
		ASSERT(MAX_PRIME_FACTOR*MAX_PRIME_FACTOR == PRIME_LIMIT);

		size_t lim = PRIME_LIMIT/2;
		dynamic_bitset<> vComposite(lim );
		for (uint32_t n=3; n<MAX_PRIME_FACTOR; n+=2) {
			if (!vComposite[n/2]) {
				for (uint32_t m=n*n/2; m<lim; m+=n)
					vComposite.set(m);
			}
		}
		push_back(2);
		for (uint32_t p=3; p<PRIME_LIMIT; p+=2)
			if (!vComposite[p/2])
				push_back(p);
	}
};

static InterlockedSingleton<PrimeNumbers> s_primeNumbers;

const vector<uint32_t>& PrimeTable() {
	return *s_primeNumbers;
}

BigInteger Primorial(uint32_t n) {
	const vector<uint32_t>& primes = PrimeTable();
	BigInteger r = 1;
	EXT_FOR (uint32_t p, primes) {
		if (p > n)
			break;
		r *= BigInteger(p);
	}
	return r;
}

}} // Ext::Num::

