#include <el/ext.h>

#if UCFG_SPECIAL_CRT
#	include "mutex"
#	include "ostream"
#	include "thread"
#endif

namespace ExtSTL  {
using namespace Ext;

#if !UCFG_STD_THREAD || (defined(_MSC_VER) && _MSC_VER < 1700) || !UCFG_STDSTL || UCFG_SPECIAL_CRT

unsigned int thread::hardware_concurrency() {
	return Environment.ProcessorCoreCount;
}

ostream& AFXAPI operator<<(ostream& os, const thread::id& v) {
	return os << v.m_tid;
}


namespace this_thread {
	
thread::id AFXAPI get_id() {
	thread::id r;
#if UCFG_USE_PTHREADS
	r.m_tid = ::pthread_self();
#elif UCFG_WDM
	r.m_tid = KeGetCurrentThread();
#else
	r.m_tid = ::GetCurrentThreadId();
#endif
	return r;
}

} // this_thread::

thread::~thread() {
	if (joinable())
		terminate();
}

thread::thread(EXT_RV_REF(thread) rv)
	:	m_t(rv.m_t)
{
	rv.m_t = nullptr;
}

thread& thread::operator=(EXT_RV_REF(thread) rv) {
	m_t = rv.m_t;
	rv.m_t = nullptr;
	return *this;
}

void thread::swap(thread& other) {
	std::swap(m_t, other.m_t);
}

void thread::detach() {
	if (!joinable())
		Ext::ThrowImp(E_INVALIDARG);
	m_t = nullptr;
}

void thread::Start(Thread *t) {
	(m_t = t)->Start();
}

void thread::join() {
	m_t->Join();
	m_t = nullptr;
}

#endif // !UCFG_STD_THREAD



}  // ExtSTL::


